@@ -1,26 +1,34 @@
 import concurrent.futures
 import threading
 import time
+from collections import defaultdict
 from copy import deepcopy
 from dataclasses import asdict
 from os import PathLike
 from pathlib import Path
 from pprint import pformat
 from tempfile import TemporaryDirectory
-from typing import Union, Sequence, Mapping, List
+from typing import Union, Sequence, Mapping, List, Tuple
 
-import pyqalx
+import OrcFxAPI
 import numpy as np
+import pyqalx
+import yaml
 from OrcFxAPI import Model
 from pyqalx import QalxSession, Item, Group, Set
+from pyqalx.core.errors import QalxEntityNotFound
+from qalx_orcaflex.gui.batch.overview_main import run_gui, BatchOverview
 
 from qalx_orcaflex import data_models
 from qalx_orcaflex.data_models import JobState
-from qalx_orcaflex.gui.batch.overview_main import run_gui, BatchOverview
 from qalx_orcaflex.helpers import (
     data_file_paths,
     required_results_from_yaml,
     clean_set_key,
+    orcaflex_file_type,
+    resolve_parents,
+    parse_restart_chain,
+    eliminate_subsets,
 )
 from qalx_orcaflex.video.helpers import unpack_model_videos, check_sim_file_type
 
@@ -60,19 +68,84 @@ class BaseSource:
         new.update(update)
         return new
 
-    def _path_meta(self, data_file: Mapping, path: PathLike) -> Path:
+
+class NewSource(BaseSource):
+    """base class for batch sources with new data files. These also need to support restarts.
+
+    Any child class must implement a `to_jobs` method which takes a base_job and
+    temp_folder argument and returns a generator of jobs
+
+    The `use_orcfxapi` keyword argument can be set false to avoid using an OrcaFlex licence when inspecting the job.
+
+    the temp_folder will be automatically deleted by the batch manager after the
+    batch has been submitted.
+    """
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.use_orcfxapi = kwargs.get("use_orcfxapi", True)
+        if self.use_orcfxapi:
+            self.model = OrcFxAPI.Model(threadCount=1)
+
+    def _link_restart_parent(self, data_file: dict, data_file_path: Path) -> None:
+        """establish the chain of restart parents for the file at ``data_file_path`` and if it exists store the
+        information in the ``data_file``
+
+        :param data_file: dict with infor on the data file
+        :param data_file_path: path of data file
+        :return: None
+        """
+        if self.use_orcfxapi:
+            self.model.LoadData(str(data_file_path))
+            chain = [str(Path(p).resolve()) for p in self.model.restartParentFileNames]
+        else:
+            chain = resolve_parents(data_file_path)
+        if chain:
+            parse_restart_chain(data_file, chain)
+
+    def _data_file_type(self, path: Path) -> OrcFxAPI.ModelType:
+        """gets the type of data OrcaFlex data file at the path.
+
+        If use_orcaflex is False then the yml will be inspected for keys that indicate the model type.
+
+        :param path:
+        :return: OrcFxAPI.ModelType
+        """
+        if self.use_orcfxapi and (path.suffix in (".yml", ".yaml")):
+            data_file_type = orcaflex_file_type(self.model, path)
+        elif (not self.use_orcfxapi) and (path.suffix in (".yml", ".yaml")):
+            with open(path, "rb") as fp:
+                yaml_data = yaml.load(fp)
+            if "RestartingFrom" in yaml_data:
+                data_file_type = OrcFxAPI.ModelType.Restart
+            elif "BaseFile" in yaml_data:
+                data_file_type = OrcFxAPI.ModelType.Variation
+            else:
+                data_file_type = OrcFxAPI.ModelType.Standard
+        else:
+            data_file_type = OrcFxAPI.ModelType.Standard
+        return data_file_type
+
+    def _inspect_file(
+        self, data_file: dict, path: PathLike
+    ) -> Tuple[Path, type(OrcFxAPI.ModelType)]:
+        """extract detail about the data file and its model type
+
+        :param data_file: dict of details about the data file
+        :param path: path to the data file
+        :return: tuple of pathlib.Path to the file and the OrcFxAPI.ModelType
+        """
         dfp = Path(path)
         data_file["meta"]["_class"] = "orcaflex.job.data_file"
         data_file["meta"]["data_file_path"] = str(dfp)
         data_file["meta"]["data_file_name"] = dfp.name
         data_file["meta"]["data_file_drive"] = dfp.drive
         data_file["meta"]["data_file_uri"] = dfp.as_uri()
-        return dfp
-
-    def to_jobs(self, base_job: Mapping, temp_folder: Path) -> Mapping:
-        raise NotImplementedError(
-            "This method needs to be defined in all BaseSource child classes"
-        )
+        data_file_type = self._data_file_type(dfp)
+        if data_file_type == OrcFxAPI.ModelType.Restart:
+            self._link_restart_parent(data_file, dfp)
+        data_file["meta"]["data_file_type"] = str(data_file_type)
+        return dfp, data_file_type
 
 
 class DataFileItemsSource(BaseSource):
@@ -98,7 +171,7 @@ class DataFileItemsSource(BaseSource):
             yield job
 
 
-class ModelSource(BaseSource):
+class ModelSource(NewSource):
     def __init__(self, model: Model, name: str):
         """an OrcFxAPI.Model instance as the source, requires a model name
 
@@ -114,68 +187,80 @@ class ModelSource(BaseSource):
         filename = f"{self.name}.dat"
         model_path = temp_folder.joinpath(filename)
         self.model.SaveData(model_path)
+
         data_file = {
             "source": model_path,
             "meta": {"_class": "orcaflex.job.data_file", "data_file_name": filename},
         }
+        if self.model.type == OrcFxAPI.ModelType.Restart:
+            chain = [str(Path(p).resolve()) for p in self.model.restartParentFileNames]
+            parse_restart_chain(data_file, chain)
         job = self._update_copy(
             base_job, {"data_file": data_file, "case_name": self._clean_name}
         )
         yield job
 
 
-class FileSource(BaseSource):
-    def __init__(self, path: PathLike):
+class FileSource(NewSource):
+    def __init__(self, path: PathLike, use_orcfxapi: bool = True):
         """a file source
 
         :param path:
+        :param use_orcfxapi: if True then will use an OrcaFlex licence to check the data_file_type and
+                             get information about restart parents.
         """
-        super(FileSource, self).__init__(path)
+        super(FileSource, self).__init__(path, use_orcfxapi=use_orcfxapi)
         self.path = path
 
-    def to_jobs(self, base_job: Mapping) -> Mapping:
+    def to_jobs(self, base_job: Mapping, temp_folder: Path) -> Mapping:
         data_file = {"source": self.path, "meta": {}}
-        dfp = self._path_meta(data_file, self.path)
+        data_file_path, data_file_type = self._inspect_file(data_file, self.path)
         job = self._update_copy(
-            base_job, {"data_file": data_file, "case_name": dfp.stem}
+            base_job,
+            {"data_file": data_file, "case_name": clean_set_key(data_file_path.stem)},
         )
         yield job
 
 
-class DirectorySource(BaseSource):
+class DirectorySource(NewSource):
     def __init__(
         self,
         directory: Union[PathLike, Sequence[PathLike]],
-        include_yaml: bool = False,
         skip_fails: bool = False,
         recursive: bool = False,
+        use_orcfxapi: bool = True,
     ):
         """takes a directory of files and adds them all.
 
         :param directory: path to search for data files
-        :param include_yaml: include yml and yaml files (default=False)
         :param skip_fails: don't error if there is a failed path e.g. it is deleted while searching (default=False)
         :param recursive: search all subdirectories (default=False)
+        :param use_orcfxapi: will make sure only OrcaFlex yaml files are added - THIS USES AN ORCAFLEX LICENCE
+                           (default=True)
         """
         super(DirectorySource, self).__init__(
-            directory, include_yaml, skip_fails, recursive
+            directory, skip_fails, recursive, use_orcfxapi=use_orcfxapi
         )
         self.directory = directory
-        self.include_yaml = include_yaml
         self.skip_fails = skip_fails
         self.recursive = recursive
+        if self.use_orcfxapi:
+            self.model = OrcFxAPI.Model(threadCount=1)
+
+    def to_jobs(self, base_job: Mapping, _) -> Mapping:
+        for dfp in data_file_paths(self.directory, self.skip_fails, self.recursive):
+            if self.use_orcfxapi and (dfp.suffix in (".yml", ".yaml")):
+                data_file_type = orcaflex_file_type(self.model, dfp)
+                if not data_file_type:
+                    continue  # this is not an OrcaFlex yml file, so we can ignore it
 
-    def to_jobs(self, base_job: Mapping, temp_folder: Path) -> Mapping:
-        for dfp in data_file_paths(
-            self.directory, self.include_yaml, self.skip_fails, self.recursive
-        ):
             data_file = {
                 "source": str(dfp),
                 "meta": {"_class": "orcaflex.job.data_file"},
             }
-            self._path_meta(data_file, dfp)
+            self._inspect_file(data_file, dfp)
             job = self._update_copy(
-                base_job, {"data_file": data_file, "case_name": dfp.stem}
+                base_job, {"data_file": data_file, "case_name": clean_set_key(dfp.stem)}
             )
             yield job
 
@@ -193,7 +278,7 @@ class OrcaFlexBatchManager:
 
     def add(
         self,
-        source: BaseSource,
+        source: Union[BaseSource, NewSource],
         job_options: data_models.OrcaFlexJobOptions = ORCA_FLEX_JOB_OPTIONS,
         required_results: Union[
             PathLike, Sequence[Union[data_models.TimeHistory, data_models.RangeGraph]]
@@ -204,7 +289,7 @@ class OrcaFlexBatchManager:
     ):
         """add data files to a batch based on a source and extra options.
 
-        :param source: a child of :ref:BaseSource that implements `to_jobs`
+        :param source: a child of :ref:BaseSource or :ref:NewSource that implements `to_jobs`
         :param job_options: various options and settings see
             :ref:data_models.OrcaFlexJobOptions
         :param required_results: an iterator containing :ref:data_models.TimeHistory
@@ -274,6 +359,7 @@ class OrcaFlexBatch:
             self.meta = standard_meta
             self.meta.update(meta)
         self.verbose = verbose
+        self.contains_restarts = None
 
     @classmethod
     def get_and_check_progress(cls, batch_name, ofx_session=None):
@@ -307,6 +393,7 @@ class OrcaFlexBatch:
                 "case_name": case_name,
                 "state": asdict(data_models.OrcaFlexJobState()),
                 "send_to": self.batch_options.send_sim_to,
+                "is_restart": False,
             },
         )
 
@@ -332,6 +419,13 @@ class OrcaFlexBatch:
                 existing_item = None
                 job_items["data_file"] = dict(**job["data_file"])
                 job["data_file"]["meta"]["case_guid"] = self.built_sets[case_name].guid
+
+            # here we need to flag a set that restarts from another simulation
+            if job["data_file"]["meta"].get("data_file_type") == "Restart":
+                self.contains_restarts = True
+                self.built_sets[case_name]["meta"]["is_restart"] = True
+            else:
+                self.built_sets[case_name]["meta"]["is_restart"] = False
         else:
             existing_item = None
 
@@ -417,9 +511,7 @@ class OrcaFlexBatch:
         if existing_item is not None:
             job_items_zipped[
                 "data_file"
-            ] = (
-                existing_item
-            )  # the existing item didn't need to be added so we add it here
+            ] = existing_item  # the existing item didn't need to be added so we add it here
         if self.verbose:  # print
             print(f"added {case_name}")
         self.session.log.info(f"created {case_name}")
@@ -462,6 +554,13 @@ class OrcaFlexBatch:
 
     def _submit(self):
         """this is the method that gets called by the BatchManager to submit the batch to a queue"""
+
+        # check if there are restarts to worry about
+        if self.contains_restarts:
+            sets_to_run = self._collate_restarts()
+            self.meta["restart_chains"] = sets_to_run
+            self.meta["contains_restarts"] = True
+
         self.session.log.info(f"Submitting batch to {self.batch_options.batch_queue}")
         self.group_batch = self.session.group.add(
             sets=self.built_sets, meta=self.meta
@@ -475,6 +574,110 @@ class OrcaFlexBatch:
         # submit batch
         Group.add_to_queue(payload=self.group_batch, queue=batch_queue)
 
+    def _collate_restarts(self) -> List:
+        """If the batch contains restarts then we need to ensure that all the parent data or simulation files are
+        available to the sim bot. This function ensures that the data or simulation files are stored in qalx for
+        any parent of any restart.
+
+        We distinguish between two types of restart in this function:
+
+        - inside: this is a parent in the batch of another data file in the batch
+        - outside: this a parent which is not included in the batch
+
+        :return: List
+        """
+        self.session.log.info("Batch contains restarts")
+        all_paths = {
+            s["items"].data_file.meta.data_file_path: s.guid
+            for n, s in self.built_sets.items()
+        }
+        restart_sets = list(
+            filter(lambda ks: ks[1].meta.is_restart, self.built_sets.items())
+        )
+
+        # If the parent of a restart is not part of the batch then we will need to do some work to make sure that the
+        # data or simulation file is stored in qalx and is referenced from the set
+
+        inside_chains = (
+            []
+        )  # chains that are completely inside the batch require further processing
+
+        for name, set_ in restart_sets:
+            if set_["items"].data_file.meta.data_file_parent not in all_paths:
+                # here we have an outside parent
+                self.session.log.info(f"{name} contains a parent outside the batch")
+                data_file_meta = set_["items"].data_file.meta
+                self.session.log.debug(
+                    "Chain of restart ancestors all being added to qalx."
+                )
+                out_chain = reversed(data_file_meta.data_file_chain)
+                # we add all of these to qalx and ensure the set has details about where to get the chain
+                chain_file_items = []
+                for ancestor in out_chain:
+                    fi = self.session.item.add(
+                        source=ancestor,
+                        meta={"_class": "orcaflex.job.parent_data_file"},
+                    )
+                    chain_file_items.append(fi.guid)
+
+                set_[
+                    "items"
+                ].data_file.meta.chain_file_items = (
+                    chain_file_items  # save the chain of restart data files
+                )
+            elif not set_["items"].data_file.meta.get("restart_parent"):
+                # these are all inside parents
+                inside_chain = set_["items"].data_file.meta.data_file_chain
+                inside_chains.append(
+                    [set_["items"].data_file.meta.data_file_path] + inside_chain
+                )
+
+            if set_["items"].data_file.meta.get(
+                "restart_parent"
+            ):  # in this case the parent sim file is available on disk
+                parent = set_["items"].data_file.meta.get("restart_parent")
+                try:  # that sim might already exist on qalx, so we attempt to retrieve it
+                    sim_item = self.session.item.find_one(
+                        query={
+                            "meta": {
+                                "_class": "orcaflex.simulation_file",
+                                "sim_hash": parent["sim_hash"],
+                            }
+                        }
+                    )
+                except QalxEntityNotFound:  # if it's not in qalx then we'll send it
+                    sim_item = self.session.item.add(
+                        source=parent["sim_path"],
+                        meta={
+                            "_class": "orcaflex.simulation_file",
+                            "sim_hash": parent["sim_hash"],
+                        },
+                    )
+                set_["items"].data_file.meta.restart_parent.sim_item = sim_item.guid
+            self.session.item.save(set_["items"].data_file)
+
+        # The code below aims to solve the problem of having large batches with complex restart hierarchies.
+        # We want to be able to run independent chains concurrently to simulate most efficiently across a distributed
+        # computing network.
+
+        # first, reduce the chains to the smallest number of complete sequences
+        maximal_chains = eliminate_subsets(inside_chains)
+
+        # then find the chains that contain common sequences
+        common = defaultdict(list)
+        for chain in maximal_chains:
+            common_elements = [not set(chain).isdisjoint(_s) for _s in maximal_chains]
+            common[tuple(common_elements)].append(chain)
+
+        # finally, flatten the chain and remove duplicates while maintaining the order
+        sets_to_run = []
+        for v in common.values():
+            flat_common_guids = [
+                all_paths[i] for s in v for i in s
+            ]  # we want a chain of UUID here rather than paths
+            sets_to_run.append(list(dict.fromkeys(flat_common_guids)))
+        return sets_to_run
+
     def __enter__(self):
         self.temp_dir.__enter__()
         self.batch_manager = OrcaFlexBatchManager(self.temp_dir.name)
